import pytest
import json
import httpx
import unittest
import datetime
from unittest.mock import ANY, AsyncMock, patch
from fastapi.testclient import TestClient
from obspy import UTCDateTime

from geomagio.api.secure import app
from geomagio.api.secure.SessionMiddleware import SessionMiddleware
from geomagio.metadata.Metadata import Metadata
from geomagio.api.db.MetadataDatabaseFactory import MetadataDatabaseFactory

client = TestClient(app)


@pytest.mark.asyncio
@patch("databases.Database.execute", new_callable=AsyncMock)
async def test_create_metadata_valid_session(
    valid_token, valid_userinfo, valid_reading_metadata, mocker
):
    valid_session = {"token": valid_token, "user": valid_userinfo}
    mock_get_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionMiddleware, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionMiddleware, "save_session", mock_save_session)

    body = json.dumps(valid_reading_metadata)

    response = client.post(
        url="/metadata", headers={"Cookie": "PHPSESSID=valid_session_id"}, content=body
    )

    assert response.status_code == 201
    mock_get_session.assert_called_once_with("valid_session_id")

    returned_metadata = Metadata(**response.json())
    assert returned_metadata.id == 1
    assert returned_metadata.category == "reading"
    assert returned_metadata.starttime == UTCDateTime("2024-11-07T12:12:12Z")
    assert returned_metadata.endtime == UTCDateTime("2024-11-07T20:20:20Z")
    assert returned_metadata.network == "NT"
    assert returned_metadata.station == "BOU"
    assert returned_metadata.status == "new"
    assert returned_metadata.priority == 1


@pytest.mark.asyncio
@patch("databases.Database.execute", new_callable=AsyncMock)
async def test_create_metadata_with_id(
    valid_token, valid_userinfo, valid_reading_metadata, mocker
):
    # input id is removed and replaced with new db id
    valid_session = {"token": valid_token, "user": valid_userinfo}
    mock_get_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionMiddleware, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionMiddleware, "save_session", mock_save_session)

    valid_reading_metadata["id"] = 1234
    body = json.dumps(valid_reading_metadata)

    response = client.post(
        url="/metadata", headers={"Cookie": "PHPSESSID=valid_session_id"}, content=body
    )

    assert response.status_code == 201
    mock_get_session.assert_called_once_with("valid_session_id")

    returned_metadata = Metadata(**response.json())
    assert returned_metadata.id == 1
    assert returned_metadata.category == "reading"
    assert returned_metadata.starttime == UTCDateTime("2024-11-07T12:12:12Z")
    assert returned_metadata.endtime == UTCDateTime("2024-11-07T20:20:20Z")
    assert returned_metadata.network == "NT"
    assert returned_metadata.station == "BOU"
    assert returned_metadata.status == "new"
    assert returned_metadata.priority == 1


@pytest.mark.asyncio
@patch("databases.Database.execute", new_callable=AsyncMock)
async def test_create_metadata_valid_auth_token(
    valid_token, valid_userinfo, valid_gitlab_user, valid_reading_metadata, mocker
):
    valid_user_response = httpx.Response(status_code=200, json=valid_gitlab_user)

    valid_groups = [
        {
            "id": 1,
            "full_path": "group1",
        },
        {"id": 2, "full_path": "group2"},
    ]

    valid_groups_response = httpx.Response(status_code=200, json=valid_groups)

    mock_gitlab_request = mocker.AsyncMock(
        side_effect=[
            valid_user_response,
            valid_groups_response,
            httpx.Response(status_code=200, json={}),
        ]
    )
    mocker.patch("httpx.AsyncClient.get", mock_gitlab_request)

    valid_session = {"token": valid_token, "user": valid_userinfo}

    mock_save_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionMiddleware, "save_session", mock_save_session)

    body = json.dumps(valid_reading_metadata)

    response = client.post(
        url="/metadata", headers={"Authorization": "auth_token"}, content=body
    )

    session_user = {
        "email": "test_email",
        "sub": 1234,
        "groups": ["group1", "group2"],
        "name": "test_user",
        "nickname": "test_user_nickname",
        "picture": "http://picture_url",
    }

    assert response.status_code == 201
    mock_save_session.assert_called_once_with(ANY, {"user": session_user})

    returned_metadata = Metadata(**response.json())
    assert returned_metadata.id == 1
    assert returned_metadata.category == "reading"
    assert returned_metadata.starttime == UTCDateTime("2024-11-07T12:12:12Z")
    assert returned_metadata.endtime == UTCDateTime("2024-11-07T20:20:20Z")
    assert returned_metadata.network == "NT"
    assert returned_metadata.station == "BOU"
    assert returned_metadata.status == "new"
    assert returned_metadata.priority == 1


@pytest.mark.asyncio
async def test_create_metadata_no_session_or_token(valid_reading_metadata, mocker):
    mock_get_session = mocker.AsyncMock()
    mocker.patch.object(SessionMiddleware, "get_session", mock_get_session)

    body = json.dumps(valid_reading_metadata)

    # do not pass in cookie or auth
    response = client.post(url="/metadata", content=body)
    assert response.status_code == 401

    mock_get_session.assert_not_called()


@pytest.mark.asyncio
async def test_create_metadata_with_no_session_and_invalid_token(mocker):
    mock_get_session = mocker.AsyncMock()
    mocker.patch.object(SessionMiddleware, "get_session", mock_get_session)

    mock_gitlab_request = mocker.AsyncMock(return_value=None)
    mocker.patch("geomagio.api.secure.login.get_gitlab_user", mock_gitlab_request)

    response = client.post(
        url="/metadata", headers={"Authorization": "invalid_gitlab_token"}
    )
    assert response.status_code == 401

    mock_get_session.assert_not_called()
    mock_gitlab_request.assert_called_once_with(token="invalid_gitlab_token")


@pytest.mark.asyncio
async def test_create_metadata_invalid_session(mocker):
    # mock invalid session. this is created when users GET /metadata without logging in
    invalid_session = {
        "redirect_uri": "redirect_uri",
        "nonce": "nonce_str",
        "url": "test_url",
    }
    mock_get_session = mocker.AsyncMock(return_value=invalid_session)
    mocker.patch.object(SessionMiddleware, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock()
    mocker.patch.object(SessionMiddleware, "save_session", mock_save_session)

    response = client.post(
        url="/metadata", headers={"Cookie": "PHPSESSID=invalid_session_id"}
    )

    assert response.status_code == 401
    mock_get_session.assert_called_once_with("invalid_session_id")


@pytest.mark.asyncio
@patch("databases.Database.transaction")
@patch("databases.Database.fetch_all", new_callable=AsyncMock)
@patch("databases.Database.execute", new_callable=AsyncMock)
async def test_update_metadata(
    mock_execute,
    mock_fetch,
    mock_transaction,
    valid_token,
    valid_userinfo,
    valid_reading_metadata,
    mocker,
):
    valid_userinfo["groups"] = ["ghsc/geomag/operations/roles/reviewer"]
    valid_session = {"token": valid_token, "user": valid_userinfo}
    mock_get_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionMiddleware, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionMiddleware, "save_session", mock_save_session)

    valid_reading_metadata["id"] = 1234
    mock_fetch.side_effect = [
        [
            {
                "id": 1234,
                "category": "instrument",
                "data_valid": True,
                "priority": 1,
            }
        ],
        [valid_reading_metadata],
    ]
    mock_execute.side_effect = [1, None]

    body = json.dumps(valid_reading_metadata)

    response = client.put(
        url="/metadata/1234",
        headers={"Cookie": "PHPSESSID=valid_session_id"},
        content=body,
    )
    assert response.status_code == 200

    returned_metadata = Metadata(**response.json())
    returned_metadata.id = 1234
    returned_metadata.category = "reading"

    assert mock_fetch.call_count == 2
    assert mock_execute.call_count == 2

    # the query for both fetch_all requests should be the same
    fetch_called_pararms = mock_fetch.call_args.args[0].compile().params
    assert fetch_called_pararms["id_1"] == 1234

    # assert save to metadata history is correct
    insert_called_params = mock_execute.call_args_list[0].args[0].compile().params
    assert insert_called_params["metadata_id"] == 1234
    assert insert_called_params["category"] == "instrument"
    assert insert_called_params["data_valid"] == True
    assert insert_called_params["priority"] == 1

    # assert update to metadata table is correct
    update_called_params = mock_execute.call_args_list[1].args[0].compile().params
    assert update_called_params["id_1"] == 1234
    assert update_called_params["category"] == "reading"


@pytest.mark.asyncio
@patch("databases.Database.transaction")
@patch("databases.Database.fetch_all", new_callable=AsyncMock)
@patch("databases.Database.execute", new_callable=AsyncMock)
async def test_update_metadata_valid_auth_token(
    mock_execute,
    mock_fetch,
    mock_transaction,
    valid_token,
    valid_userinfo,
    valid_gitlab_user,
    valid_reading_metadata,
    mocker,
):
    valid_user_response = httpx.Response(status_code=200, json=valid_gitlab_user)

    valid_groups = [
        {
            "id": 1,
            "full_path": "group1",
        },
        {"id": 2, "full_path": "ghsc/geomag/operations/roles/reviewer"},
    ]

    valid_groups_response = httpx.Response(status_code=200, json=valid_groups)

    mock_gitlab_request = mocker.AsyncMock(
        side_effect=[
            valid_user_response,
            valid_groups_response,
            httpx.Response(status_code=200, json={}),
        ]
    )
    mocker.patch("httpx.AsyncClient.get", mock_gitlab_request)

    valid_session = {"token": valid_token, "user": valid_userinfo}

    mock_save_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionMiddleware, "save_session", mock_save_session)

    valid_reading_metadata["id"] = 1234
    mock_fetch.side_effect = [
        [
            {
                "id": 1234,
                "category": "instrument",
                "data_valid": True,
                "priority": 1,
            }
        ],
        [valid_reading_metadata],
    ]
    mock_execute.side_effect = [1, None]

    body = json.dumps(valid_reading_metadata)

    response = client.put(
        url="/metadata/1234", headers={"Authorization": "auth_token"}, content=body
    )

    session_user = {
        "email": "test_email",
        "sub": 1234,
        "groups": ["group1", "ghsc/geomag/operations/roles/reviewer"],
        "name": "test_user",
        "nickname": "test_user_nickname",
        "picture": "http://picture_url",
    }

    assert response.status_code == 200
    mock_save_session.assert_called_once_with(ANY, {"user": session_user})

    returned_metadata = Metadata(**response.json())
    returned_metadata.id = 1234
    returned_metadata.category = "reading"

    assert mock_fetch.call_count == 2
    assert mock_execute.call_count == 2

    # the query for both fetch_all requests should be the same
    fetch_called_pararms = mock_fetch.call_args.args[0].compile().params
    assert fetch_called_pararms["id_1"] == 1234

    # assert save to metadata history is correct
    insert_called_params = mock_execute.call_args_list[0].args[0].compile().params
    assert insert_called_params["metadata_id"] == 1234
    assert insert_called_params["category"] == "instrument"
    assert insert_called_params["data_valid"] == True
    assert insert_called_params["priority"] == 1

    # assert update to metadata table is correct
    update_called_params = mock_execute.call_args_list[1].args[0].compile().params
    assert update_called_params["id_1"] == 1234
    assert update_called_params["category"] == "reading"


@pytest.mark.asyncio
@patch("databases.Database.transaction")
@patch("databases.Database.fetch_all", new_callable=AsyncMock)
@patch("databases.Database.execute", new_callable=AsyncMock)
async def test_update_metadata_no_existing_metadata(
    mock_execute,
    mock_fetch,
    mock_transaction,
    valid_token,
    valid_userinfo,
    valid_reading_metadata,
    mocker,
):
    valid_userinfo["groups"] = ["ghsc/geomag/operations/roles/reviewer"]
    valid_session = {"token": valid_token, "user": valid_userinfo}
    mock_get_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionMiddleware, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionMiddleware, "save_session", mock_save_session)

    valid_reading_metadata["id"] = 1234
    mock_fetch.return_value = []

    body = json.dumps(valid_reading_metadata)

    try:
        response = client.put(
            url="/metadata/1234",
            headers={"Cookie": "PHPSESSID=valid_session_id"},
            content=body,
        )
    except Exception as e:
        assert type(e) == ValueError


@pytest.mark.asyncio
async def test_update_metadata_no_valid_group(
    valid_token, valid_userinfo, valid_reading_metadata, mocker
):
    valid_session = {"token": valid_token, "user": valid_userinfo}
    mock_get_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionMiddleware, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionMiddleware, "save_session", mock_save_session)

    body = json.dumps(valid_reading_metadata)

    response = client.put(
        url="/metadata/1234",
        headers={"Cookie": "PHPSESSID=valid_session_id"},
        content=body,
    )
    assert response.status_code == 403


@pytest.mark.asyncio
async def test_update_metadata_no_session(
    valid_token, valid_userinfo, valid_reading_metadata, mocker
):
    valid_session = {"token": valid_token, "user": valid_userinfo}
    mock_get_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionMiddleware, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionMiddleware, "save_session", mock_save_session)

    body = json.dumps(valid_reading_metadata)

    response = client.put(url="/metadata/1234", content=body)
    assert response.status_code == 401

    mock_get_session.assert_not_called()


@pytest.mark.asyncio
async def test_update_metadata_invalid_session(mocker, valid_reading_metadata):
    # mock invalid session. this is created when users GET /metadata without logging in
    invalid_session = {
        "redirect_uri": "redirect_uri",
        "nonce": "nonce_str",
        "url": "test_url",
    }
    mock_get_session = mocker.AsyncMock(return_value=invalid_session)
    mocker.patch.object(SessionMiddleware, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock()
    mocker.patch.object(SessionMiddleware, "save_session", mock_save_session)

    body = json.dumps(valid_reading_metadata)

    response = client.put(
        url="/metadata/1234",
        headers={"Cookie": "PHPSESSID=invalid_session_id"},
        content=body,
    )

    assert response.status_code == 401
    mock_get_session.assert_called_once_with("invalid_session_id")


@pytest.mark.asyncio
async def test_update_metadata_with_no_session_and_invalid_token(
    mocker, valid_reading_metadata
):
    mock_get_session = mocker.AsyncMock()
    mocker.patch.object(SessionMiddleware, "get_session", mock_get_session)

    mock_gitlab_request = mocker.AsyncMock(return_value=None)
    mocker.patch("geomagio.api.secure.login.get_gitlab_user", mock_gitlab_request)

    body = json.dumps(valid_reading_metadata)

    response = client.put(
        url="/metadata/1234",
        headers={"Authorization": "invalid_gitlab_token"},
        content=body,
    )
    assert response.status_code == 401

    mock_get_session.assert_not_called()
    mock_gitlab_request.assert_called_once_with(token="invalid_gitlab_token")


# # # the way the database class is set up prevents us from using pytest-mocks here. unittest is a bit more
# # # powerful but less intuitive and allows us to inspect the parameters to mocked database calls
class TestMetadata(unittest.IsolatedAsyncioTestCase):
    @patch("databases.Database.fetch_all", new_callable=AsyncMock)
    async def test_get_metadata(self, mock_fetch):
        mock_fetch.return_value = [
            {
                "id": 8372,
                "station": "BOU",
                "category": "instrument",
                "network": "NT",
                "channel": "F",
                "location": "R0",
                "data_valid": True,
                "status": "new",
                "metadata": {"metadata1": "metadata1"},
            },
            {
                "id": 8376,
                "station": "BOU",
                "category": "instrument",
                "network": "NT",
                "channel": "F",
                "location": "R0",
                "data_valid": True,
                "status": "new",
                "metadata": {"metadata2": "metadata2"},
            },
        ]

        response = client.get(
            url="/metadata?station=BOU&category=instrument&network=NT&channel=F&location=R0&data_valid=true&status=new"
        )
        assert response.status_code == 200

        mock_fetch.assert_called_once()
        called_pararms = mock_fetch.call_args.args[0].compile().params

        assert called_pararms["station_1"] == "BOU"
        assert called_pararms["category_1"] == "instrument"
        assert called_pararms["network_1"] == "NT"
        assert called_pararms["channel_1"] == "F"
        assert called_pararms["location_1"] == "R0"
        assert called_pararms["status_1"] == ["new"]

        returned_metadata = []

        rows = response.json()
        for metadata in rows:
            returned_metadata.append(Metadata(**metadata))

        assert returned_metadata[0].id == 8372
        assert returned_metadata[1].id == 8376

    @patch("databases.Database.fetch_all", new_callable=AsyncMock)
    async def test_get_metadata_with_starttime(self, mock_fetch):
        mock_fetch.return_value = [
            {
                "id": 8372,
                "station": "BOU",
                "category": "instrument",
                "network": "NT",
                "channel": "F",
                "location": "R0",
                "data_valid": True,
                "status": "new",
                "metadata": {"metadata1": "metadata1"},
            },
            {
                "id": 8376,
                "station": "BOU",
                "category": "instrument",
                "network": "NT",
                "channel": "F",
                "location": "R0",
                "data_valid": True,
                "status": "new",
                "metadata": {"metadata2": "metadata2"},
            },
        ]
        response = client.get(
            url="/metadata?station=BOU&category=instrument&starttime=2024-11-06T04:27:40Z"
        )
        assert response.status_code == 200

        mock_fetch.assert_called_once()
        called_pararms = mock_fetch.call_args.args[0].compile().params

        assert called_pararms["station_1"] == "BOU"
        assert called_pararms["category_1"] == "instrument"
        # assert endtime is set to starttime because the query is where endtime is None or greater than starttime
        assert called_pararms["endtime_1"] == datetime.datetime(
            2024, 11, 6, 4, 27, 40, tzinfo=datetime.timezone.utc
        )

        returned_metadata = []

        rows = response.json()
        for metadata in rows:
            returned_metadata.append(Metadata(**metadata))

        assert returned_metadata[0].id == 8372
        assert returned_metadata[1].id == 8376

    @patch("databases.Database.fetch_all", new_callable=AsyncMock)
    async def test_get_metadata_with_endtime(self, mock_fetch):
        mock_fetch.return_value = [
            {
                "id": 8372,
                "station": "BOU",
                "category": "instrument",
                "network": "NT",
                "channel": "F",
                "location": "R0",
                "data_valid": True,
                "status": "new",
                "metadata": {"metadata1": "metadata1"},
            },
            {
                "id": 8376,
                "station": "BOU",
                "category": "instrument",
                "network": "NT",
                "channel": "F",
                "location": "R0",
                "data_valid": True,
                "status": "new",
                "metadata": {"metadata2": "metadata2"},
            },
        ]
        response = client.get(
            url="/metadata?station=BOU&category=instrument&endtime=2024-12-06T04:27:40Z"
        )
        assert response.status_code == 200

        mock_fetch.assert_called_once()
        called_pararms = mock_fetch.call_args.args[0].compile().params

        assert called_pararms["station_1"] == "BOU"
        assert called_pararms["category_1"] == "instrument"
        # assert starttime is set to endtime because the query is where starttime is None or greater than endtime
        assert called_pararms["starttime_1"] == datetime.datetime(
            2024, 12, 6, 4, 27, 40, tzinfo=datetime.timezone.utc
        )

        returned_metadata = []

        rows = response.json()
        for metadata in rows:
            returned_metadata.append(Metadata(**metadata))

        assert returned_metadata[0].id == 8372
        assert returned_metadata[1].id == 8376

    @patch("databases.Database.fetch_all", new_callable=AsyncMock)
    async def test_get_metadata_with_created_after(self, mock_fetch):
        mock_fetch.return_value = [
            {
                "id": 8372,
                "station": "BOU",
                "category": "instrument",
                "network": "NT",
                "channel": "F",
                "location": "R0",
                "data_valid": True,
                "status": "new",
                "metadata": {"metadata1": "metadata1"},
            },
            {
                "id": 8376,
                "station": "BOU",
                "category": "instrument",
                "network": "NT",
                "channel": "F",
                "location": "R0",
                "data_valid": True,
                "status": "new",
                "metadata": {"metadata2": "metadata2"},
            },
        ]
        response = client.get(
            url="/metadata?station=BOU&category=instrument&created_after=2024-10-06T04:27:40Z"
        )
        assert response.status_code == 200

        mock_fetch.assert_called_once()
        called_pararms = mock_fetch.call_args.args[0].compile().params

        assert called_pararms["station_1"] == "BOU"
        assert called_pararms["category_1"] == "instrument"
        # assert created_time is set to created_after because the query is where created_time is greater than created_after
        assert called_pararms["created_time_1"] == datetime.datetime(
            2024, 10, 6, 4, 27, 40, tzinfo=datetime.timezone.utc
        )

        returned_metadata = []

        rows = response.json()
        for metadata in rows:
            returned_metadata.append(Metadata(**metadata))

        assert returned_metadata[0].id == 8372
        assert returned_metadata[1].id == 8376

    @patch("databases.Database.fetch_all", new_callable=AsyncMock)
    async def test_get_metadata_with_created_before(self, mock_fetch):
        mock_fetch.return_value = [
            {
                "id": 8372,
                "station": "BOU",
                "category": "instrument",
                "network": "NT",
                "channel": "F",
                "location": "R0",
                "data_valid": True,
                "status": "new",
                "metadata": {"metadata1": "metadata1"},
            },
            {
                "id": 8376,
                "station": "BOU",
                "category": "instrument",
                "network": "NT",
                "channel": "F",
                "location": "R0",
                "data_valid": True,
                "status": "new",
                "metadata": {"metadata2": "metadata2"},
            },
        ]
        response = client.get(
            url="/metadata?station=BOU&category=instrument&created_before=2024-09-06T04:27:40Z"
        )
        assert response.status_code == 200

        mock_fetch.assert_called_once()
        called_pararms = mock_fetch.call_args.args[0].compile().params

        assert called_pararms["station_1"] == "BOU"
        assert called_pararms["category_1"] == "instrument"
        # assert created_time is set to created_before because the query is where created_time is less than created_before
        assert called_pararms["created_time_1"] == datetime.datetime(
            2024, 9, 6, 4, 27, 40, tzinfo=datetime.timezone.utc
        )

        returned_metadata = []

        rows = response.json()
        for metadata in rows:
            returned_metadata.append(Metadata(**metadata))

        assert returned_metadata[0].id == 8372
        assert returned_metadata[1].id == 8376

    @patch("databases.Database.fetch_all", new_callable=AsyncMock)
    async def test_get_metadata_returns_empty(self, mock_fetch):
        mock_fetch.return_value = []

        response = client.get(
            url="/metadata?station=BOU&category=instrument&created_before=2024-09-06T04:27:40Z"
        )
        assert response.status_code == 200

        mock_fetch.assert_called_once()

        # assert response is an empty list
        assert response.json() == []

    @patch("databases.Database.fetch_all", new_callable=AsyncMock)
    async def test_get_metadata_by_id(self, mock_fetch):
        mock_fetch.return_value = [{"id": 1234, "category": "instrument"}]
        response = client.get(url="/metadata/1234")
        assert response.status_code == 200

        mock_fetch.assert_called_once()
        called_pararms = mock_fetch.call_args.args[0].compile().params

        assert called_pararms["id_1"] == 1234

        returned_metadata = Metadata(**response.json())
        assert returned_metadata.id == 1234
        assert returned_metadata.category == "instrument"

    @patch("databases.Database.fetch_all", new_callable=AsyncMock)
    async def test_get_metadata_by_id_returns_empty(self, mock_fetch):
        mock_fetch.return_value = []
        try:
            client.get(url="/metadata/1234")
        except Exception as e:
            assert type(e) == ValueError

        mock_fetch.assert_called_once()

    @patch("databases.Database.fetch_one", new_callable=AsyncMock)
    async def test_get_metadata_history_by_id(self, mock_fetch):
        mock_fetch.return_value = {"id": 1234, "category": "instrument"}
        response = client.get(url="/metadata/history/1234")
        assert response.status_code == 200

        mock_fetch.assert_called_once()
        called_pararms = mock_fetch.call_args.args[0].compile().params

        assert called_pararms["id_1"] == 1234

        returned_metadata = Metadata(**response.json())
        assert returned_metadata.id == 1234
        assert returned_metadata.category == "instrument"

    @patch("databases.Database.fetch_one", new_callable=AsyncMock)
    async def test_get_metadata_history_by_id_returns_empty(self, mock_fetch):
        mock_fetch.return_value = None

        response = client.get(url="/metadata/history/1234")
        assert response.status_code == 404

        mock_fetch.assert_called_once()
        called_pararms = mock_fetch.call_args.args[0].compile().params

        assert called_pararms["id_1"] == 1234

    @patch("databases.Database.transaction")
    @patch("databases.Database.fetch_all", new_callable=AsyncMock)
    async def test_get_metadata_history_by_metadata_id(
        self, mock_fetch, mock_transaction
    ):
        mock_fetch.side_effect = [
            [{"id": 242, "metadata_id": 1234, "category": "instrument"}],
            [{"id": 1234, "category": "reading"}],
        ]
        response = client.get(url="/metadata/1234/history")
        assert response.status_code == 200

        assert mock_fetch.call_count == 2
        called_pararms = mock_fetch.call_args.args[0].compile().params

        assert called_pararms["id_1"] == 1234

        returned_metadata = []

        rows = response.json()
        for metadata in rows:
            returned_metadata.append(Metadata(**metadata))

        # expect the metadata from the metadata table to be first
        assert returned_metadata[0].id == 1234
        assert returned_metadata[1].id == 242

    @patch("databases.Database.transaction")
    @patch("databases.Database.fetch_all", new_callable=AsyncMock)
    async def test_get_metadata_history_by_metadata_id_no_history(
        self, mock_fetch, mock_transaction
    ):
        mock_fetch.side_effect = [[], [{"id": 1234, "category": "reading"}]]
        response = client.get(url="/metadata/1234/history")
        assert response.status_code == 200

        assert mock_fetch.call_count == 2
        called_pararms = mock_fetch.call_args.args[0].compile().params

        assert called_pararms["id_1"] == 1234

        returned_metadata = []

        rows = response.json()
        for metadata in rows:
            returned_metadata.append(Metadata(**metadata))

        assert returned_metadata[0].id == 1234

    @patch("databases.Database.transaction")
    @patch("databases.Database.fetch_all", new_callable=AsyncMock)
    async def test_get_metadata_history_by_metadata_id_no_history_or_metadata(
        self, mock_fetch, mock_transaction
    ):
        mock_fetch.side_effect = [[], []]

        try:
            client.get(url="/metadata/1234/history")
        except Exception as e:
            assert type(e) == ValueError

        assert mock_fetch.call_count == 2
